#include <glib.h>
#include <math.h>

#include "gui.h"
#include "output.h"

typedef struct _Point Point;
struct _Point {
  double x, y;
};

typedef struct _Vector Vector;
struct _Vector {
  double x, y;
};

typedef struct _Model Model;
typedef struct _Object Object;
typedef struct _Attractor Attractor;

#define MAX_ATTRACTORS 5

struct _Attractor {
  Object *object;
  Vector offset;
};

struct _Object {
  int num_attractors;
  Attractor attractors[MAX_ATTRACTORS];

  Vector attractor_offset;

  Point position;
  Point next_position;
  Vector velocity;
  Vector acceleration;
  double mass;
  double theta;
};

struct _Model {
  int num_objects;
  Object *objects;
  Object attractor;

  Object *current_attractor;
  double k;		/* Spring constant */
  double friction;	/* Friction constant */

  double last_time;
  double steps;
};


static void
model_move_attractor (Model *model, double x, double y)
{
  model->current_attractor->position.x = x;
  model->current_attractor->position.y = y;
}

static void
model_init_object (Object *object,
		   double position_x, double position_y,
		   double velocity_x, double velocity_y, double mass)
{
  object->num_attractors = 0;

  object->position.x = position_x;
  object->position.y = position_y;

  object->velocity.x = velocity_x;
  object->velocity.y = velocity_y;

  object->acceleration.x = 0;
  object->acceleration.y = 0;

  object->mass = mass;
}

static void
model_add_object_attractor (Object *object, Object *attractor_object,
			    double offset_x, double offset_y)
{
  Attractor *attractor;

  attractor = &object->attractors[object->num_attractors];
  object->num_attractors++;

  attractor->object = attractor_object;
  attractor->offset.x = offset_x;
  attractor->offset.y = offset_y;
}


static void
model_set_grid_size (Model *model, int width, int height,
		     int grid_width, int grid_height)
{
  int grid_x, grid_y, i;
  const double hpad = (double) width / 3;
  const double vpad = (double) height / 3;

  i = 0;
  for (grid_y = 0; grid_y < grid_height; grid_y++)
    for (grid_x = 0; grid_x < grid_width; grid_x++) {
      model->objects[i].num_attractors = 0;
      if (grid_x > 0)
	model_add_object_attractor (&model->objects[i],
				    &model->objects[i - 1],
				    hpad, 0);

      if (grid_y > 0)
	model_add_object_attractor (&model->objects[i],
				    &model->objects[i - grid_width],
				    0, vpad);
      i++;
    }
}

static void
model_init_grid (Model *model, int x, int y, int width, int height,
		 int grid_width, int grid_height)
{
  int grid_x, grid_y, i;
  const double hpad = (double) width / 3;
  const double vpad = (double) height / 3;

  model->num_objects = grid_width * grid_height;
  model->objects = g_new (Object, model->num_objects);

  i = 0;
  for (grid_y = 0; grid_y < grid_height; grid_y++)
    for (grid_x = 0; grid_x < grid_width; grid_x++) {
      model_init_object (&model->objects[i],
#if 1
			 x, y,
#else
			 x + grid_x * width / (grid_width - 1),
			 y + grid_y * height / (grid_height - 1),
#endif
			 -50, -50, 20);
      
#if 0
      if (i == 0)
	model_add_object_attractor (&model->objects[i],
				    model->current_attractor,
				    0, 0);
#endif

#define CX 0
#define CY 0

#if 0
      if (grid_x < CX)
	model_add_object_attractor (&model->objects[i],
				    &model->objects[i + 1],
				    -hpad, 0);
      if (grid_x > CX)
	model_add_object_attractor (&model->objects[i],
				    &model->objects[i - 1],
				    hpad, 0);

      if (grid_y < CY)
	model_add_object_attractor (&model->objects[i],
				    &model->objects[i + grid_width],
				    0, -vpad);
      if (grid_y > CY)
	model_add_object_attractor (&model->objects[i],
				    &model->objects[i - grid_width],
				    0, vpad);
#endif

#if 1
      if (grid_x > 0)
	model_add_object_attractor (&model->objects[i],
				    &model->objects[i - 1],
				    hpad, 0);

      if (grid_y > 0)
	model_add_object_attractor (&model->objects[i],
				    &model->objects[i - grid_width],
				    0, vpad);
#endif
#if 0
      if (grid_x < grid_width - 1)
	model_add_object_attractor (&model->objects[i],
				    &model->objects[i + 1],
				    -hpad, 0);

      if (grid_y < grid_height - 1)
	model_add_object_attractor (&model->objects[i],
				    &model->objects[i + grid_width],
				    0, -vpad);
#endif

      i++;
    }
}

static Model *
model_new (int x, int y, int width, int height)
{
  Model *model;

  model = g_new (Model, 1);

  model->current_attractor = &model->attractor;
  model_move_attractor (model, 200, 140);
  model->k = 1.6;
  model->friction = 5.0;
  model->last_time = 0;
  model->steps = 0;

  model_init_grid (model, x, y, width, height, 4, 4);

  return model;
}

static void
model_step_object (Model *model, Object *object)
{
  Vector target;
  Vector force;
  int i;

  force.x = 0;
  force.y = 0;
  for (i = 0; i < object->num_attractors; i++) {
    target.x = object->attractors[i].object->position.x +
      object->attractors[i].offset.x;
    target.y = object->attractors[i].object->position.y +
      object->attractors[i].offset.y;

    force.x += model->k * (target.x - object->position.x) -
      model->friction * object->velocity.x;
    force.y += model->k * (target.y - object->position.y) - 
      model->friction * object->velocity.y;
  }

  object->acceleration.x = force.x / object->mass;
  object->acceleration.y = force.y / object->mass;

  object->velocity.x += object->acceleration.x;
  object->velocity.y += object->acceleration.y;

  object->next_position.x = object->position.x + object->velocity.x;
  object->next_position.y = object->position.y + object->velocity.y;

  object->theta += 0.05;
}

static int
model_step (Model *model, double t)
{
  int i, j, steps;

  model->steps += (t - model->last_time) / 0.03;
  model->last_time = t;
  steps = floor(model->steps);
  model->steps -= steps;

  for (j = 0; j < steps; j++)
    {
      for (i = 0; i < model->num_objects; i++)
	model_step_object (model, &model->objects[i]);

      for (i = 0; i < model->num_objects; i++) {
	if (&model->objects[i] == model->current_attractor)
	  continue;
	model->objects[i].position = model->objects[i].next_position;
      }
    }

  return j;
}

static void
bezier_patch_evaluate (Model *model, double u, double v,
		       double *patch_x, double *patch_y)
{
  double coeffs_u[4], coeffs_v[4];
  double x, y;
  int i, j;
  
  coeffs_u[0] = (1 - u) * (1 - u) * (1 - u);
  coeffs_u[1] = 3 * u * (1 - u) * (1 - u);
  coeffs_u[2] = 3 * u * u * (1 - u);
  coeffs_u[3] = u * u * u;

  coeffs_v[0] = (1 - v) * (1 - v) * (1 - v);
  coeffs_v[1] = 3 * v * (1 - v) * (1 - v);
  coeffs_v[2] = 3 * v * v * (1 - v);
  coeffs_v[3] = v * v * v;

  x = 0;
  y = 0;
  for (i = 0; i < 4; i++)
    for (j = 0; j < 4; j++)
      {
	x += coeffs_u[i] * coeffs_v[j] * model->objects[j * 4 + i].position.x;
	y += coeffs_u[i] * coeffs_v[j] * model->objects[j * 4 + i].position.y;
      }

  *patch_x = x;
  *patch_y = y;
}

static void
wobble (int u, int v,
	int x, int y, int width, int height,
	int *deformed_x, int *deformed_y,
	void *p)
{
  Model *model = p;
  double patch_x, patch_y;

  bezier_patch_evaluate (model,
			 (double) u / width, (double) v / height,
			 &patch_x, &patch_y);

  *deformed_x = patch_x + 0.5;
  *deformed_y = patch_y + 0.5;

}

static gboolean
wobbly_callback (LmcAnimation *animation, gpointer user_data)
{
  Model *model = user_data;
  LmcOutputWindow *owindow;
  gboolean first;
  double t;

  owindow = g_object_get_data (G_OBJECT (animation->toplevel),
			       "gui-output-window");
  if (owindow == NULL)
    return FALSE;

  first = (model->last_time == 0);
  t = g_timer_elapsed (animation->timer, NULL);
  if (model_step (model, t) == 0 && !first)
    return TRUE;

  lmc_output_freeze (animation->output);
  
  lmc_output_window_set_deformation (owindow, wobble, model);

  lmc_output_thaw (animation->output);

  return TRUE;
}

void
destroy_animation (void *data)
{
  LmcAnimation *animation = data;

  lmc_animation_unref (animation);
}

static void
on_position_changed (LmcToplevel *toplevel,
		     Model       *model)
{
  LmcBorderInfo border_info;
  int x, y;
  
  lmc_toplevel_get_border_info (toplevel, &border_info);
  lmc_toplevel_get_position (toplevel, &x, &y);
  model_move_attractor (model, x - border_info.left, y - border_info.top);
}

static void
on_size_changed (LmcToplevel *toplevel,
		 Model       *model)
{
  int x, y, width, height;
  LmcBorderInfo border_info;

  lmc_toplevel_get_position (toplevel, &x, &y);
  lmc_toplevel_get_size (toplevel, &width, &height);
  lmc_toplevel_get_border_info (toplevel, &border_info);
  width += border_info.left + border_info.right;
  height += border_info.top + border_info.bottom;
  
  model_set_grid_size (model, width, height, 4, 4);
}

void
lmc_wobble_setup (LmcToplevel *toplevel,
		  LmcGui      *gui)
{
  Model *model;
  int x, y, width, height;
  LmcBorderInfo border_info;

  lmc_toplevel_get_set_position (toplevel, &x, &y);
  lmc_toplevel_get_size (toplevel, &width, &height);
  lmc_toplevel_get_border_info (toplevel, &border_info);

  width += border_info.left + border_info.right;
  height += border_info.top + border_info.bottom;

  model = model_new (x, y, width, height);
  model->current_attractor = &model->objects[0];
  model_move_attractor (model, x - border_info.left, y - border_info.top);

  g_signal_connect (toplevel, "position-changed",
		    G_CALLBACK (on_position_changed), model);
  g_signal_connect (toplevel, "size-changed",
		    G_CALLBACK (on_size_changed), model);

  lmc_animation_new (gui, toplevel, wobbly_callback, model);
}
